/**
 *  joola.io
 *
 *  Copyright Joola Smart Solutions, Ltd. <info@joo.la>
 *
 *  Licensed under GNU General Public License 3.0 or later.
 *  Some rights reserved. See LICENSE, AUTHORS.
 *
 *  @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
 */

var
  logger = require('joola.io.logger'),

  _datasources = require('./lib/objects/datasources'),

  _ = global._ = require('underscore'),
  ce = global.ce = require('cloneextend'),

  express = require('express'),
  http = require('http'),
  https = require('https'),
  path = require('path'),
  nconf = require('nconf'),
  async = require('async');

require('nconf-http');
require('pkginfo')(module, 'version');

var joola = {};
global.joola = joola;

joola.state = {
  status: 'init',
  running: false
};

joola.logger = logger;

joola.config = nconf;
joola.config.server = null;
joola.config.auth = null;
joola.config.integration = null;
joola.config.content = null;

var
  httpServer,
  httpsServer,
  app = global.app = express();

var fetchConfig = function (name, file, callback) {
  joola.config.add(name, { type: 'http', url: 'http://localhost:40001/conf/' + file, callback: function (err) {
    if (err) {
      //Fallback to file
      joola.config.add(name, { type: 'file', file: joola.config.get('conf') || './config/' + file + '.json' });
    }
    //Validate config
    if (!joola.config.get(name + ':version'))
      return callback(new Error('Failed to load configuration file: ' + './config/' + file + '.json'));

    return callback(null);
  }});
};

var loadConfig = function (callback) {
  joola.state.status = 'loadConfig';
  joola.config.argv()
    .env();

  //Load integration config
  //server
  fetchConfig('server', 'joola.io.engine.server', function (err) {
    if (err)
      throw err;
    joola.config.server = joola.config.get('server');
    //auth
    fetchConfig('auth', 'joola.io.engine.auth', function (err) {
      if (err)
        throw err;
      joola.config.auth = joola.config.get('auth');
      //integration
      fetchConfig('integration', 'joola.io.engine.integration', function (err) {
        if (err)
          throw err;
        joola.config.integration = joola.config.get('integration');
        //content
        fetchConfig('content', 'joola.io.engine.content', function (err) {
          if (err)
            throw err;
          joola.config.content = joola.config.get('content');
          return callback(null);
        });
      });
    });
  });
};

var setupCache = function (callback) {
  var port, host, options;

  host = joola.config.get('server:redis:host');
  port = joola.config.get('server:redis:port');

  joola.redis = require("redis");
  joola.redis.client = joola.redis.createClient();

  require('./lib/shared/cache').cache(joola, function (err) {
    if (err)
      throw err;

    joola.logger.info('Connected to Redis @ ' + host + ':' + port);

    callback(null);
  });
};

var processIntegration = function (callback) {
  joola.state.status = 'processIntegration';
  var datasources = joola.config.integration.datasources;
  datasources.forEach(function (ds) {
    if (!ds.id)
      throw 'JSON Config Validation Error: Source ' + ds.name + ' has no id';

    _.each(ds.datatables, function (dt) {
      if (!dt.id)
        throw 'JSON Config Validation Error: Table ' + dt.name + ' has no id';

      dt.datasourceid = ds.id;
    });
  });

  return callback(null);
};

var verifyDatasources = function (callback) {
  var datasources = joola.config.integration.datasources;

  var checks = [];

  datasources.forEach(function (ds) {
    var check = function (callback) {
      joola.logger.debug('Checking data source [' + ds.id + ']...');
      _datasources.validate(ds.id, function (err, validated) {
        if (err)
          joola.logger.warn('Failed to validate data source: ' + ds.id);
        else if (!validated)
          joola.logger.warn('Failed to validate data source: ' + ds.id);

        return callback();
      });
    };
    checks.push(check);
  });

  async.parallel(checks, function () {
    return callback();
  });
};

var preFlight = function (callback) {
  joola.state.status = 'preFlight';
  loadConfig(function (err) {
    if (err)
      throw err;
    joola.logger.info('Loaded server config version ' + joola.config.get('server:version'));
    joola.logger.info('Loaded auth config version ' + joola.config.get('auth:version'));
    joola.logger.info('Loaded integration config version ' + joola.config.get('integration:version'));
    joola.logger.info('Loaded content config version ' + joola.config.get('content:version'));

    joola.logger.setLevel(joola.config.server.logLevel);

    setupCache(function (err) {
      if (err)
        throw err;
      processIntegration(function (err) {
        if (err)
          throw err;

        verifyDatasources(function () {
          callback();
        });
      });
    });
  });
};

var start = function (callback) {
  joola.state.status = 'starting';
  joola.state.running = false;

  setupApplication(function (err) {
    if (err)
      callback(err);

    setupRoutes(function (err) {
      if (err)
        callback(err);

      setupControlPort(function () {
        joola.logger.info('Control port running on port ' + joola.config.get('server:controlPort:port'));

        startHTTP(function () {
          joola.logger.debug('HTTP running');

          if (joola.config.get('server:secure') === true) {
            startHTTPS(function () {
              joola.logger.debug('HTTPS running');

              return callback(null);
            });
          }
          else
            return callback(null);
        });
      });
    });
  });
};

var setupApplication = function (callback) {
  var winstonStream = {
    write: function (message, encoding) {
      joola.logger.info(message);
    }
  };
  app.use(express.logger((global.test ? function (req, res) {
  } : {stream: winstonStream})));

  app.set('views', __dirname + '/views');
  app.set('view engine', 'jade');
  app.use(express.compress());
  app.use(express.bodyParser());
  app.use(express.methodOverride());

  callback();
};

var setupRoutes = function (callback) {
  var
    index = require('./routes/index');

  app.configure(function () {
    app.use(function (req, res, next) {
      res.setHeader('Access-Control-Allow-Origin', '*');
      res.setHeader('Access-Control-Allow-Methods', 'GET, POST, DELETE, OPTIONS');
      res.setHeader('Access-Control-Allow-Headers', 'joola-token, Content-Type, origin');
      res.setHeader('Access-Control-Max-Age', '86400');
      res.setHeader('X-Powered-By', 'joola.io');

      return next();
    });
  });

  app.get('/', index.index);
  app.get('/:resource', index.route);
  app.get('/:resource/:action', index.route);
  app.use(express.static(path.join(__dirname, 'public')));
  app.use(app.router);


  //TODO: Setup 500 and 404 routes
  callback();
};


var startHTTP = function (callback) {
  var result = {};
  try {
    var _httpServer = http.createServer(app).listen(joola.config.get('server:port'),function (err) {
      if (err) {
        result.status = 'Failed: ' + ex.message;
        return callback(result);
      }
      joola.state.status = 'running';
      joola.state.running = true;

      joola.logger.info('Joola Analytics HTTP server listening on port ' + joola.config.get('server:port'));
      httpServer = _httpServer;
      return callback(result);
    }).on('error',function (ex) {
        result.status = 'Failed: ' + ex.message;
        return callback(result);
      }).on('close', function () {

        joola.state.status = 'stopped';
        joola.state.running = false;

        joola.logger.warn('Joola Analytics HTTP server listening on port ' + (joola.config.get('server:port')).toString() + ' received a CLOSE command.');
      });
  }
  catch (ex) {
    result.status = 'Failed: ' + ex.message;
    return callback(result);
  }
  return null;
};

var startHTTPS = function (callback) {
  var result = {};
  try {
    var secureOptions = {
      key: fs.readFileSync(joola.config.get('server:keyFile')),
      cert: fs.readFileSync(joola.config.get('server:certFile'))
    };
    var _httpsServer = https.createServer(secureOptions, app).listen(joola.config.get('server:securePort'),function (err) {
      if (err) {
        result.status = 'Failed: ' + ex.message;
        return callback(result);
      }
      joola.logger.info('Joola Analytics HTTPS server listening on port ' + joola.config.get('server:securePort'));
      result.status = 'Success';
      httpsServer = _httpsServer;
      return callback(result);
    }).on('error',function (ex) {
        result.status = 'Failed: ' + ex.message;
        return callback(result);
      }).on('close', function () {
        joola.logger.warn('Joola Analytics HTTPS server listening on port ' + joola.config.get('server:securePort').toString() + ' received a CLOSE command.');
      });
  }
  catch (ex) {
    result.status = 'Failed: ' + ex.message;
    return callback(result);
  }
  return null;
};

//Control Port
var setupControlPort = function (callback) {
  var cp = require('node-controlport');
  var cp_endpoints = [];

  cp_endpoints.push({
    endpoint: 'status',
    exec: function (callback) {
      callback({status: joola.state.status, pid: process.pid});
    }
  });

  cp_endpoints.push({
      endpoint: 'start',
      exec: function (callback) {
        if (joola.config.get('server:secure') === true) {
          startHTTP(function () {
            startHTTPS(callback);
          });
        }
        else {
          startHTTP(callback);
        }
      }
    }
  );

  cp_endpoints.push({
    endpoint: 'stop',
    exec: function (callback) {
      var result = {};
      result.status = 'Success';
      try {
        httpServer.close();
        if (joola.config.get('server:secure') === true)
          httpsServer.close();

        process.exit(0);
      }
      catch (ex) {
        console.log(ex);
        result.status = 'Failed: ' + ex.message;
        return callback(result);
      }
      return callback(result);
    }
  });

  cp.start(joola.config.get('server:controlPort:port'), cp_endpoints, callback);
};

joola.logger.info('Starting joola.io Engine, version ' + module.exports.version + '...');
preFlight(function () {
  joola.logger.info('Preflight checks done, starting engine...');
  start(function (err) {
    if (err)
      throw err;

    joola.logger.info('joole.io Engine started!');
  });
});
