/**
 *  joola.io
 *
 *  Copyright Joola Smart Solutions, Ltd. <info@joo.la>
 *
 *  Licensed under GNU General Public License 3.0 or later.
 *  Some rights reserved. See LICENSE, AUTHORS.
 *
 *  @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
 */

var
  _datatables = require('../objects/datatables'),
  connector = require('../connectors/connector'),
  utils = require('../shared/utils');

exports.login = function (authToken, ip, username, password, next) {
  var _self = this;

  var auth = joola.config.auth;
  var store = auth.store;

  joola.logger.info('SSO Authenticating [' + username + '] with store [' + store + '], from [' + ip + ']...');

  if (auth.whitelist.indexOf(ip) == -1) {
    joola.logger.error('SSO Authentication failed, whitelist violation [' + ip + ']!');
    next(null);
  }

  if (auth.authToken != authToken) {
    joola.logger.error('SSO Authentication failed, wrong auth token passed [' + authToken + ']!');
    next(null);
  }

  var authenticator;

  if (store == 'joola')store = 'internal';

  switch (store) {
    case 'internal':
      authenticator = require('./joola');
      break;
    default:
      authenticator = require('./' + store);
      break;
  }

  authenticator.login(username, password, function (user) {
    if (user) {
      _self.generateToken(user, function (user, token) {
        next(user, token);
      });
    }
    else
      next(null);
  });
};

exports.checkTokenExist = function (token, next) {
  var result = false;

  joola.cache.load('tokens', token, function (err, value) {
    if (value)
      next(token, value);
    else
      next(null);
  });

  return result;
};

exports.validateToken = function (token, next) {
  var _self = this;
  _self.checkTokenExist(token, function (exists, cachedValue) {
    if (exists) {
      joola.cache.save('tokens', token, cachedValue, joola.config.auth.tokenExpiresAfter, function (err) {
        if (err)
          throw new Error('Failed to validate token [' + user.displayName + ']:' + err);

        joola.logger.info('Extended security token [' + token + '] for user [' + cachedValue.displayName + '].');

      });
      next(cachedValue, token);
    }
    else {
      next(null);
    }
  });
};

exports.generateToken = function (user, next) {
  var _self = this;
  var token = utils.shorten();//require('node-uuid').v4();

  _self.checkTokenExist(token, function (exists) {
    if (exists)
      throw new Error('Failed to generate token, one already exists [' + token + '].');

    joola.cache.save('tokens', token, user, joola.config.auth.tokenExpiresAfter, function (err) {
      if (err)
        throw new Error('Failed to save token [' + user.displayName + ']:' + err);

      joola.logger.info('Generated security token [' + token + '] for user [' + user.displayName + '].');
      next(user, token);
    })
  });
};

exports.expireToken = function (token, next) {
  joola.cache.destroy(token, next);
};

exports.logout = function (connection, callback) {
  api.session.delete(connection, function () {
    callback();
  })
};

exports.checkLoggedIn = function (connection, callback) {
  api.session.checkAuth(connection, function (session) {
    session.actionCounter++;
    callback(session, true);
  }, function () {
    callback(null, false);
  });
};

exports.hasRole = function (availableRoles, existingRoles) {
  var exist = _.find(availableRoles, function (available) {
    return _.find(existingRoles, function (existing) {
      return existing == available;
    })
  });

  return exist != null;
};