/**
 *  joola.io
 *
 *  Copyright Joola Smart Solutions, Ltd. <info@joo.la>
 *
 *  Licensed under GNU General Public License 3.0 or later.
 *  Some rights reserved. See LICENSE, AUTHORS.
 *
 *  @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
 */

var
  joola = global.joola, //for jslint
  utils = require('../shared/utils'),
  _datasources = require('../objects/datasources'),
  _datatables = require('../objects/datatables'),
  connector = require('../connectors/connector'),

  ce = require('cloneextend'),
  _mongo = require('mongodb');

exports.minCacheDate = function (datatable, callback) {
  joola.logger.debug('Checking min cache date on [' + datatable.id + ']...');
  var connstring = datatable.caching.system.url;
  joola.logger.silly('Connecting to mongo @ ' + connstring + '...');

  //noinspection JSPotentiallyInvalidConstructorUsage
  var mongo = new _mongo.MongoClient.connect(connstring, {w: 0}, function (err, db) {
    db.createCollection(datatable.id, null, function (err, collection) {
      if (err)
        throw err;
      collection.find().sort({date: 1}).limit(1).toArray(function (err, result) {
        if (err)
          throw err;
        try {
          callback(new Date(result[0].date));
        }
        catch (ex) {
          joola.logger.error('Failed to fetch min cache date: ' + ex.message);
          callback(null);
        }
        finally {
          db.close();
        }
      });
    });
  });
};

exports.maxCacheDate = function (datatable, callback) {
  var connstring = datatable.caching.system.url;
  joola.logger.silly('Connecting to mongo @ ' + connstring + '...');

  //noinspection JSPotentiallyInvalidConstructorUsage
  var mongo = new _mongo.MongoClient.connect(connstring, {w: 0}, function (err, db) {
    db.createCollection(datatable.id, null, function (err, collection) {
      if (err) throw err;
      collection.find().sort({date: -1}).limit(1).toArray(function (err, result) {
        try {
          callback(new Date(result[0].date));
        }
        catch (ex) {
          joola.logger.error('Failed to fetch max cache date: ' + ex.message);
          callback(null);
        }
        finally {
          db.close();
        }
      });
    });
  });
};

exports.maxNotHandledCacheDate = function (datatable, callback) {
  joola.logger.debug('Checking max not handled cache date on [' + datatable.id + ']...');
  var connstring = datatable.caching.system.url;
  joola.logger.silly('Connecting to mongo @ ' + connstring + '...');

  //noinspection JSPotentiallyInvalidConstructorUsage
  var mongo = new _mongo.MongoClient.connect(connstring, {w: 0}, function (err, db) {
    db.createCollection(datatable.id, null, function (err, collection) {
      if (err) throw err;
      collection.find({handled: false}).sort({date: -1}).limit(1).toArray(function (err, result) {
        db.close();
        try {
          if (result && result.length > 0)
            callback(new Date(result[0].date));
          else
            callback(null);
        }
        catch (ex) {
          callback(null);
        }
        db.close();
      });
    });
  });
};

exports.minNotHandledCacheDate = function (datatable, callback) {
  joola.logger.debug('Checking min not handled cache date on [' + datatable.id + ']...');
  var connstring = datatable.caching.system.url;
  joola.logger.silly('Connecting to mongo @ ' + connstring + '...');

  //noinspection JSPotentiallyInvalidConstructorUsage
  var mongo = new _mongo.MongoClient.connect(connstring, {w: 0}, function (err, db) {
    db.createCollection(datatable.id, null, function (err, collection) {
      if (err)
        throw err;
      collection.find({handled: false}).sort({date: 1}, null, null).limit(1).toArray(function (err, result) {
        try {
          if (result && result.length > 0)
            callback(new Date(result[0].date));
          else
            callback(null);
        }
        catch (ex) {
          throw ex;
        }
        db.close();
      });
    });
  });
};

function verifyCollection(db, datatable, query, data, callback) {
  //noinspection JSUnresolvedFunction
  joola.logger.silly('Creating collection [' + datatable.id + ']...');
  db.createCollection(datatable.id, null, function (err, collection) {
    if (err) throw err;

    //noinspection JSUnresolvedFunction
    joola.logger.silly('...Created collection [' + datatable.id + ']');

    var indexKey = {key: 1};
    verifyIndex(indexKey, collection, datatable, query, data, callback);
  });
}

function verifyIndex(indexKey, collection, datatable, query, data, callback) {
  var _indexKey = JSON.stringify(indexKey).replace(/"/g, '').replace(/:/g, '').replace(/ /g, '').replace(/\./g, '').replace(/,/g, '').replace('{', '').replace('}', '');
  joola.logger.debug('Verifying index [' + _indexKey + ']...');
  collection.ensureIndex(indexKey, { unique: true }, function (err) {
    joola.logger.debug('... Index verfieid [' + _indexKey + ']' + (err ? ' Error: ' + err : ''));

    indexKey = {'date': 1};
    collection.ensureIndex(indexKey, { unique: false }, function () {
      processRows(collection, datatable, query, data, callback);
    });
  });
}

function processDocument(datatable, chunk, documents, callback) {
  var dimensions = _datatables.basequery(datatable).dimensionsUsed;
  var metrics = _datatables.basequery(datatable).metricsUsed;

  var rowIndex = 1;
  _.each(chunk, function (row) {
      var document = {};
      document.date = new Date(row['date']).fixDate(true, true);

      var _date = ce.clone(document.date);

      var bucket = {};
      _date.setMilliseconds(0);
      bucket['second'] = new Date(_date);
      _date.setSeconds(0);
      bucket['minute'] = new Date(_date);
      _date.setMinutes(0);
      bucket['hour'] = new Date(_date);
      _date.setUTCHours(0, 0, 0, 0);
      bucket['day'] = new Date(_date);
      _date.setDate(1);
      bucket['month'] = new Date(_date);
      _date.setMonth(1);
      bucket['year'] = new Date(_date);

      document.timebucket = bucket;

      var rowKey = {date: document.date };
      document.handled = false;
      _.each(dimensions, function (d) {
        var _name = (d.alias ? d.alias : d.name);
        if (row[_name] == '')
          row[_name] = '(not set)';
        document[_name] = row[_name];
        rowKey[_name] = row[_name];
      });
      _.each(metrics, function (m) {
        var _name = m.name;
        try {
          var value = parseFloat(row[_name]);
          if (isNaN(value))
            value = null;
          document[_name] = value;
        }
        catch (ex) {
          console.log('ex: ' + ex.message);
        }
      });

      document.key = JSON.stringify(rowKey).clean();
      document.rowIndex = rowIndex;
      documents.push(document);

      rowIndex++;
    }
  )  ;
  callback();
}

function processRows(collection, datatable, query, data, callback) {
  //noinspection JSUnresolvedFunction
  joola.logger.silly('Processing rows, total count: ' + data.rows.length);

  if (data.rows.length == 0) {
    joola.logger.debug('Nothing to do here.');
    callback();
    return;
  }

  var chunks = chunk(data.rows, datatable.caching.chunkSize);
  joola.logger.debug('Broke ' + data.rows.length + ' rows into ' + chunks.length + ' chunks of data [chunkSize:' + datatable.caching.chunkSize + '].');

  data = null;

  var documents = [];
  var $options = {};

  var callsb = [];
  _.each(chunks, function (chunk) {
    joola.logger.debug('Processing chunk...');
    var callb = function (callback) {
      processDocument(datatable, chunk, documents, callback);
    };
    callsb.push(callb);
  });
  chunks = null;

  fork(callsb, function () {
    callsb = null;
    joola.logger.debug('Finished parsing documents.');
    //callback(null);

    var documentChunks = chunk(documents, datatable.caching.chunkSize);
    joola.logger.debug('Broke ' + documents.length + ' documents into ' + documentChunks.length + ' chunks of data [chunkSize:' + datatable.caching.chunkSize + '].');
    var calls = [];
    _.each(documentChunks, function (chunk) {
      var call = function (callback) {
        collection.insert(
          chunk,
          $options,
          function (err) {
            if (err) {
              joola.logger.warn('Found a duplicate record in mongodb store. Iterating through documents...: ' + err);

              var callsb = [];

              _.each(chunk, function (document) {
                var insertDocument = function (callback) {
                  joola.logger.silly('Saving document [' + document.key + '].');
                  collection.insert(document, $options, function (err) {
                    joola.logger.silly('Document saved [' + document.key + '].');
                    if (err) {
                      if (err.message.indexOf('E11000') != -1) {
                        //check if we need to update docs in this case.

                      }
                      else
                        joola.logger.error('Failed to save document [' + document.key + ']: ' + err);
                    }
                    callback();
                  });
                };
                callsb.push(insertDocument);
              });

              fork(callsb, function () {
                //delete callsb;
                joola.logger.debug('...Saved all non-duplicate documents.');
                callback();
              });
            }
            else {
              joola.logger.debug('Documents chunk saved (' + chunk.length + ' rows).');
              callback();
            }
          });
      };
      calls.push(call);
    });
    documentChunks = null;
    fork(calls, function () {
      calls = null;

      exports.minNotHandledCacheDate(datatable, function (firstDate) {
        exports.maxNotHandledCacheDate(datatable, function (lastDate) {
          //TODO: If we extend on both sides, we should not mark as handled the entire lot.
          //crunch(datatable, query, collection, db, function () {
          //   joola.logger.debug('Finished saving source data [' + firstDate + ']-[' + lastDate + '].');

          if (!query.balanced) {
            collection.update(
              {handled: false},//, date: {$lte: $enddate, $gte: $startdate}},
              {$set: {handled: true}},
              {w: 1, multi: true, verbose: true},
              function (err) {
                if (err)
                  throw err;
                joola.logger.debug('Processed rows marked as handled.');
                try {
                  if (firstDate) {
                    joola.logger.debug('First date: ' + firstDate);
                    //firstDate.fixDate(true, true);
                    //joola.logger.debug('Fixed first date: ' + firstDate);
                  }
                  else
                    joola.logger.warn('Missing first date');
                }
                catch (ex) {
                  joola.logger.warn('Failed to fixdate on firstdate: ' + firstDate + ', ex:' + ex.message);
                }
                try {
                  if (lastDate) {
                    joola.logger.debug('Last date: ' + lastDate);
                    //lastDate.fixDate(true, true);
                    //joola.logger.debug('Fixed last date: ' + lastDate);
                  }
                  else
                    joola.logger.warn('Missing last date');
                }
                catch (ex) {
                  joola.logger.warn('Failed to fixdate on lastdate: ' + lastDate + ', ex:' + ex.message);
                }
                callback(null, documents, firstDate, lastDate);
              });
          }
          else
            callback(null, documents, firstDate, lastDate);
          //});
        });
      });
    })
  });
}

exports.saveData = function (query, datatable, data, callback) {
  var connstring = datatable.caching.system.url;
  joola.logger.silly('Connecting to mongo @ ' + connstring + '...');

  //noinspection JSPotentiallyInvalidConstructorUsage
  var mongo = new _mongo.MongoClient.connect(connstring, {w: 0}, function (err, db) {
    if (err) throw err;

    joola.logger.debug('Connected to mongo @ ' + connstring);

    verifyCollection(db, datatable, query, data, function (err, documents, firstDate, lastDate) {
      db.close();
      callback(err, documents, firstDate, lastDate);
    });
  });
};

exports.fetch = function (query, callback) {
  var connstring = query.datatable.caching.system.url;
  joola.logger.silly('Connecting to mongo @ ' + connstring + '...');
  //noinspection JSPotentiallyInvalidConstructorUsage
  var mongo = new _mongo.MongoClient.connect(connstring, {w: 0}, function (err, db) {
    if (err) throw err;

    joola.logger.silly('Verifying collection [' + query.datatable.id + '.' + query.resolution + ']...');
    db.createCollection(query.datatable.id, {strict: false}, function (err, collection) {
      var processFetch = function (collection) {
        var fields = {};
        _.each(query.dimensions, function (d) {
          fields[d.name] = 1;
        });

        joola.logger.debug('Executing fetch from [' + query.datatable.id + '], range [' + utils.formatDate(query.startdate, 'yyyy-mm-dd hh:nn:ss') + ']-[' + utils.formatDate(query.enddate, 'yyyy-mm-dd hh:nn:ss') + ']...');

        var ds = _datasources.get(query.datatable.datasourceid);
        var dbquery = connector.createQuery();

        query.datatable.datasource = ds;
        query.datatable.query = _datatables.basequery(query.datatable);

        var basequery = query.datatable.query;
        dbquery.sql = basequery.sql;
        dbquery.datasource = ds;

        dbquery.enddate = new Date(query.enddate).fixDate();
        dbquery.startdate = new Date(query.startdate).fixDate();
        dbquery.resolution = query.resolution;
        dbquery._query = query;

        try {
          var manager = require('./manager'); //avoid circular requirement
          //manager.cacheTable(query.datatable, ce.clone(dbquery), function () {
          var groupBy = {};
          _.each(query.dimensions, function (d) {
            if (d.type == 'date') {
              switch (query.resolution) {
                case 'second':
                  groupBy['date'] = '$timebucket.second';
                  break;
                case 'minute':
                  groupBy['date'] = '$timebucket.minute';
                  break;
                case 'hour':
                  groupBy['date'] = '$timebucket.hour';
                  break;
                case 'day':
                  groupBy['date'] = '$timebucket.day';
                  break;
                case 'week':
                  groupBy['date'] = '$timebucket.day';
                  break;
                case 'month':
                  groupBy['date'] = '$timebucket.month';
                  break;
                case 'year':
                  groupBy['date'] = '$timebucket.year';
                  break;
                default:
                  break;
              }
            }
            else
              groupBy[d.name] = '$' + d.name;
          });

          var $group = {};
          $group._id = groupBy;

          if (!query.distinctCount) {
            $group.rowcount = {$sum: 1};

            _.each(query.metrics, function (m) {
              var exist = _.find(query.datatable.metrics, function (m) {
                return _.find(query.metrics, function (m2) {
                  return m2.id == m.id;
                })
              });
              if (exist) {
                $group[m.name] = {$sum: '$' + m.name};
              }
            });
          }
          else {

          }

          var $sortkey = {};

          if (query.sortKey) {
            if ($group[query.sortKey.name])
              $sortkey[query.sortKey.name] = (query.sortDir == 'DESC' ? -1 : 1);
            else {
              $sortkey['_id.' + query.sortKey.name] = (query.sortDir == 'DESC' ? -1 : 1);
            }
          }
          else {
            if (query.distinctCount) {
              if (Object.keys($group._id).length > 1) {
                if ($group._id.date)
                  $sortkey['_id.date'] = (query.sortDir == 'DESC' ? -1 : 1);
                else
                  $sortkey['_id'] = (query.sortDir == 'DESC' ? -1 : 1);
              }
              else
                $sortkey['_id'] = (query.sortDir == 'DESC' ? -1 : 1);
            }
            else {
              $sortkey['_id'] = (query.sortDir == 'DESC' ? -1 : 1);
            }
          }

          var $match = {
            date: { $gte: new Date(query.startdate), $lte: new Date(query.enddate)}
          };

          _.each(query.filters, function (filter) {
            if (filter.dimension.type != 'date') {
              if (filter.operator == '=')
                $match[filter.dimension.name] = filter.value;
              else if (filter.operator == '*=')
                $match[filter.dimension.name] = eval('/' + filter.value + '/gi');
              else
                $match[filter.dimension.name] = filter.value;
            }
          });

          $match['handled'] = true;


          if (!query.distinctCount) {
            //console.log('g', $group);
            //console.log('sort', $sortkey);
            joola.logger.debug('Executing mongodb aggregate [' + query.startdate + ']-[' + query.enddate + ']...');
            collection.aggregate(
              {$match: $match},
              {$group: $group},
              {$sort: $sortkey},
              //{$limit: 500},
              function (err, result) {
                if (!err) {
                  joola.logger.debug('Found [' + result.length + ' documents] in cached collection [' + query.datatable.id + '.' + query.resolution + ']');

                  function countKeys(obj) {
                    return Object.keys(obj).length;
                  }

                  if (result.length == 1) {
                    var r = result[0];
                    if (countKeys(r._id) == 0) {
                      joola.logger.warn('Found a missing dimension in collection.');
                      r._id = $group._id;
                      _.each(r._id, function (value, key) {
                        r._id[key] = '(not set)';
                      });
                    }
                  }

                  callback(err, result, query);
                }
                else
                  callback(err, null, query);
                db.close();
              });
          }
          else {
            joola.logger.debug('Executing mongodb distinct aggregate [' + query.startdate + ']-[' + query.enddate + ']...');
            groupBy = {};
            $group = {};
            var $unwind = {};
            _.each(query.dimensions, function (d) {
              if (d.type == 'date') {
                switch (query.resolution) {
                  case 'second':
                    groupBy['date'] = '$timebucket.second';
                    break;
                  case 'minute':
                    groupBy['date'] = '$timebucket.minute';
                    break;
                  case 'hour':
                    groupBy['date'] = '$timebucket.hour';
                    break;
                  case 'day':
                    groupBy['date'] = '$timebucket.day';
                    break;
                  case 'week':
                    groupBy['date'] = '$timebucket.day';
                    break;
                  case 'month':
                    groupBy['date'] = '$timebucket.month';
                    break;
                  case 'year':
                    groupBy['date'] = '$timebucket.year';
                    break;
                  default:
                    break;
                }
              }
              else if (d.id != query.metrics[0].dimension) {
                groupBy[d.name] = '$' + d.name;
              }
              else {
                $group[d.name] = {'$addToSet': '$' + d.name};
                $unwind = '$' + d.name;
              }
            });
            $group._id = groupBy;
            //

            var $group2 = {
              _id: '$_id'
            };
            $group2[query.metrics[0].name] = {$sum: 1};

            //console.log('m', $match);
            //console.log('g1', $group);
            //console.log('un', $unwind);
            //console.log('g2', $group2);
            $sortkey = {};

            if (query.sortKey) {
              if ($group[query.sortKey.name])
                $sortkey[query.sortKey.name] = (query.sortDir == 'DESC' ? -1 : 1);
              else {
                $sortkey[ query.sortKey.name] = (query.sortDir == 'DESC' ? -1 : 1);
              }
            }
            else {
              if (query.distinctCount) {
                if (Object.keys($group._id).length > 1) {
                  if ($group._id.date)
                    $sortkey['_id.date'] = (query.sortDir == 'DESC' ? -1 : 1);
                  else
                    $sortkey['_id'] = (query.sortDir == 'DESC' ? -1 : 1);
                }
                else
                  $sortkey['_id'] = (query.sortDir == 'DESC' ? -1 : 1);
              }
              else {
                $sortkey['_id'] = (query.sortDir == 'DESC' ? -1 : 1);
              }
            }

            //console.log('g', $group);
            //console.log('u', $unwind);
            //console.log('g2', $group2);
            //console.log('sort', $sortkey);

            collection.aggregate(
              {$match: $match},
              {$group: $group},
              {$unwind: $unwind},
              {$group: $group2},
              {$sort: $sortkey},
              //{$limit: 500},
              function (err, result) {

                if (!err) {
                  joola.logger.debug('Found [' + result.length + ' documents] in cached collection [' + query.datatable.id + '.' + query.resolution + ']');

                  // console.log(result);

                  function countKeys(obj) {
                    return Object.keys(obj).length;
                  }

                  if (result.length == 1) {
                    var r = result[0];
                    if (countKeys(r._id) == 0) {
                      joola.logger.warn('Found a missing dimension in collection.');
                      r._id = $group._id;
                      _.each(r._id, function (value, key) {
                        r._id[key] = '(not set)';
                      })
                    }
                  }

                  callback(err, result, query);
                }
                else {
                  callback(err, null, query);
                }
                db.close();
              }
            );
          }
          //});
        }
        catch (ex) {
          joola.logger.error('Error while calling cacheTable: ' + ex.message);
          throw ex;
        }
      };

      if (err) {
        db.collection(query.datatable.id, function (err, collection) {
          if (err)
            throw err;
          processFetch(collection);
        });
      }
      else {
        processFetch(collection);
      }
    });
  });
};